from typing import Any, Dict, Set


class GlobalState(object):
    """
    This class is just a placeholder to share data amongst different
    versions of Nornir after running ``filter`` multiple times.

    Attributes:
        failed_hosts: Hosts that have failed to run a task properly
    """

    __slots__ = "dry_run", "failed_hosts"

    def __init__(self, dry_run: bool = False, failed_hosts: Set[str] = None) -> None:
        self.dry_run = dry_run
        self.failed_hosts = failed_hosts or set()

    def recover_host(self, host: str) -> None:
        """Remove ``host`` from list of failed hosts."""
        self.failed_hosts.discard(host)

    def reset_failed_hosts(self) -> None:
        """Reset failed hosts and make all hosts available for future tasks."""
        self.failed_hosts = set()

    def dict(self) -> Dict[str, Any]:
        """ Return a dictionary representing the object. """
        return {item: getattr(self, item) for item in GlobalState.__slots__}



from typing import List
from concurrent.futures import ThreadPoolExecutor

from nornir.core.task import AggregatedResult, Task
from nornir.core.inventory import Host


class SerialRunner:
    """
    SerialRunner runs the task over each host one after the other without any parellelization
    """

    def __init__(self) -> None:
        pass

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        result = AggregatedResult(task.name)
        for host in hosts:
            result[host.name] = task.copy().start(host)
        return result


class ThreadedRunner:
    """
    ThreadedRunner runs the task over each host using threads

    Arguments:
        num_workers: number of threads to use
    """

    def __init__(self, num_workers: int = 20) -> None:
        self.num_workers = num_workers

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        result = AggregatedResult(task.name)
        futures = []
        with ThreadPoolExecutor(self.num_workers) as pool:
            for host in hosts:
                future = pool.submit(task.copy().start, host)
                futures.append(future)

        for future in futures:
            worker_result = future.result()
            result[worker_result.host.name] = worker_result
        return result



import logging
import pathlib
from typing import Any, Dict, Type

from nornir.core.inventory import (
    Inventory,
    Group,
    Groups,
    Host,
    Hosts,
    Defaults,
    ConnectionOptions,
    HostOrGroup,
    ParentGroups,
)

import ruamel.yaml

logger = logging.getLogger(__name__)


def _get_connection_options(data: Dict[str, Any]) -> Dict[str, ConnectionOptions]:
    cp = {}
    for cn, c in data.items():
        cp[cn] = ConnectionOptions(
            hostname=c.get("hostname"),
            port=c.get("port"),
            username=c.get("username"),
            password=c.get("password"),
            platform=c.get("platform"),
            extras=c.get("extras"),
        )
    return cp


def _get_defaults(data: Dict[str, Any]) -> Defaults:
    return Defaults(
        hostname=data.get("hostname"),
        port=data.get("port"),
        username=data.get("username"),
        password=data.get("password"),
        platform=data.get("platform"),
        data=data.get("data"),
        connection_options=_get_connection_options(data.get("connection_options", {})),
    )


def _get_inventory_element(
    typ: Type[HostOrGroup], data: Dict[str, Any], name: str, defaults: Defaults
) -> HostOrGroup:
    return typ(
        name=name,
        hostname=data.get("hostname"),
        port=data.get("port"),
        username=data.get("username"),
        password=data.get("password"),
        platform=data.get("platform"),
        data=data.get("data"),
        groups=data.get(
            "groups"
        ),  # this is a hack, we will convert it later to the correct type
        defaults=defaults,
        connection_options=_get_connection_options(data.get("connection_options", {})),
    )


class SimpleInventory:
    def __init__(
        self,
        host_file: str = "hosts.yaml",
        group_file: str = "groups.yaml",
        defaults_file: str = "defaults.yaml",
    ) -> None:
        """
        SimpleInventory is an inventory plugin that loads data from YAML files.
        The YAML files follow the same structure as the native objects

        Args:

          host_file: path to file with hosts definition
          group_file: path to file with groups definition. If
                it doesn't exist it will be skipped
          defaults_file: path to file with defaults definition.
                If it doesn't exist it will be skipped
        """

        self.host_file = pathlib.Path(host_file).expanduser()
        self.group_file = pathlib.Path(group_file).expanduser()
        self.defaults_file = pathlib.Path(defaults_file).expanduser()

    def load(self) -> Inventory:
        yml = ruamel.yaml.YAML(typ="safe")

        if self.defaults_file.exists():
            with open(self.defaults_file, "r") as f:
                defaults_dict = yml.load(f) or {}
            defaults = _get_defaults(defaults_dict)
        else:
            defaults = Defaults()

        hosts = Hosts()
        with open(self.host_file, "r") as f:
            hosts_dict = yml.load(f)

        for n, h in hosts_dict.items():
            hosts[n] = _get_inventory_element(Host, h, n, defaults)

        groups = Groups()
        if self.group_file.exists():
            with open(self.group_file, "r") as f:
                groups_dict = yml.load(f) or {}

            for n, g in groups_dict.items():
                groups[n] = _get_inventory_element(Group, g, n, defaults)

            for h in hosts.values():
                h.groups = ParentGroups([groups[g] for g in h.groups])

            for g in groups.values():
                g.groups = ParentGroups([groups[g] for g in g.groups])

        return Inventory(hosts=hosts, groups=groups, defaults=defaults)



from typing import List

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task

from typing_extensions import Protocol


class Processor(Protocol):
    """
    This defines the Processor interface. A processor plugin needs to implement each method with the
    same exact signature. It's not necessary to subclass it.

    A processor is a plugin that gets called when certain events happen.
    """

    def task_started(self, task: Task) -> None:
        """
        This method is called right before starting the task
        """
        raise NotImplementedError("needs to be implemented by the processor")

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        """
        This method is called when all the hosts have completed executing their respective task
        """
        raise NotImplementedError("needs to be implemented by the processor")

    def task_instance_started(self, task: Task, host: Host) -> None:
        """
        This method is called before a host starts executing its instance of the task
        """
        raise NotImplementedError("needs to be implemented by the processor")

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        """
        This method is called when a host completes its instance of a task
        """
        raise NotImplementedError("needs to be implemented by the processor")

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        """
        This method is called before a host starts executing a subtask
        """
        raise NotImplementedError("needs to be implemented by the processor")

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        """
        This method is called when a host completes executing a subtask
        """
        raise NotImplementedError("needs to be implemented by the processor")


class Processors(List[Processor]):
    """
    Processors is a wrapper class that holds a list of Processor. Each method
    will just iterate over all the the Processor objects in ``self`` and call
    its method. For instance::

        >>>    def my_method(...):
        >>>        for p in self:
        >>>            p.my_method(...)
    """

    def task_started(self, task: Task) -> None:
        for p in self:
            p.task_started(task)

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        for p in self:
            p.task_completed(task, result)

    def task_instance_started(self, task: Task, host: Host) -> None:
        for p in self:
            p.task_instance_started(task, host)

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        for p in self:
            p.task_instance_completed(task, host, result)

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        for p in self:
            p.subtask_instance_started(task, host)

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        for p in self:
            p.subtask_instance_completed(task, host, result)



from typing import Any, Type

from nornir.core.inventory import Inventory, TransformFunction
from nornir.core.plugins.register import PluginRegister

from typing_extensions import Protocol


INVENTORY_PLUGIN_PATH = "nornir.plugins.inventory"
TRANSFORM_FUNCTION_PLUGIN_PATH = "nornir.plugins.transform_function"


class InventoryPlugin(Protocol):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        This method configures the plugin
        """
        raise NotImplementedError("needs to be implemented by the plugin")

    def load(self) -> Inventory:
        """
        This method implements the plugin's business logic
        """
        raise NotImplementedError("needs to be implemented by the plugin")


InventoryPluginRegister: PluginRegister[Type[InventoryPlugin]] = PluginRegister(
    INVENTORY_PLUGIN_PATH
)


TransformFunctionRegister: PluginRegister[TransformFunction] = PluginRegister(
    TRANSFORM_FUNCTION_PLUGIN_PATH
)



from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Union,
    KeysView,
    ValuesView,
    ItemsView,
    Iterator,
    TypeVar,
)

from nornir.core.configuration import Config
from nornir.core.plugins.connections import (
    ConnectionPlugin,
    ConnectionPluginRegister,
)
from nornir.core.exceptions import ConnectionAlreadyOpen, ConnectionNotOpen

from mypy_extensions import Arg, KwArg


HostOrGroup = TypeVar("HostOrGroup", "Host", "Group")


class BaseAttributes(object):
    __slots__ = ("hostname", "port", "username", "password", "platform")

    def __init__(
        self,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        platform: Optional[str] = None,
    ) -> None:
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.platform = platform

    @classmethod
    def schema(cls) -> Dict[str, Any]:
        return {
            "hostname": "str",
            "port": "int",
            "username": "str",
            "password": "str",
            "platform": "str",
        }

    def dict(self) -> Dict[str, Any]:
        return {
            "hostname": object.__getattribute__(self, "hostname"),
            "port": object.__getattribute__(self, "port"),
            "username": object.__getattribute__(self, "username"),
            "password": object.__getattribute__(self, "password"),
            "platform": object.__getattribute__(self, "platform"),
        }


class ConnectionOptions(BaseAttributes):
    __slots__ = ("extras",)

    def __init__(
        self,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        platform: Optional[str] = None,
        extras: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.extras = extras
        super().__init__(
            hostname=hostname,
            port=port,
            username=username,
            password=password,
            platform=platform,
        )

    @classmethod
    def schema(self) -> Dict[str, Any]:
        return {
            "extras": {"$key": "$value"},
            **super().schema(),
        }

    def dict(self) -> Dict[str, Any]:
        return {
            "extras": self.extras,
            **super().dict(),
        }


class ParentGroups(List["Group"]):
    def __contains__(self, value: object) -> bool:
        if isinstance(value, str):
            return any([value == g.name for g in self])
        else:
            return any([value == g for g in self])

    def add(self, group: "Group") -> None:
        """
        Add the ParentGroup.  The group will only be appended
        if it does not exist.

        :param group: Parent Group object to add
        :return: None
        """
        # only add the group if it doesn't exist
        if not self.__contains__(group):
            self.append(group)


class InventoryElement(BaseAttributes):
    __slots__ = ("groups", "data", "connection_options")

    def __init__(
        self,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        platform: Optional[str] = None,
        groups: Optional[ParentGroups] = None,
        data: Optional[Dict[str, Any]] = None,
        connection_options: Optional[Dict[str, ConnectionOptions]] = None,
    ) -> None:
        self.groups = groups or ParentGroups()
        self.data = data or {}
        self.connection_options = connection_options or {}
        super().__init__(
            hostname=hostname,
            port=port,
            username=username,
            password=password,
            platform=platform,
        )

    @classmethod
    def schema(self) -> Dict[str, Any]:
        return {
            "groups": ["$group_name"],
            "data": {"$key": "$value"},
            "connection_options": {"$connection_type": ConnectionOptions.schema()},
            **super().schema(),
        }

    def dict(self) -> Dict[str, Any]:
        return {
            "groups": [g.name for g in self.groups],
            "data": self.data,
            "connection_options": {
                k: v.dict() for k, v in self.connection_options.items()
            },
            **super().dict(),
        }

    def extended_groups(self) -> List["Group"]:
        """
        Returns the groups this host belongs to by virtue of inheritance.

        This list is ordered based on the inheritance rules and groups are not
        duplicated. For instance, given a host with the following groups:

        hostA:
            groups:
                - group_a
                - group_b

        group_a:
            groups:
                - group_1
                - group_2
        group_b:
            groups:
                - group_2
                - group_3

        group_1:
            groups:
                - group_X

        this will return [group_a, group_1, group_X, group_2, group_b, group_3]
        """
        groups: List["Group"] = []

        for g in self.groups:
            if g not in groups:
                groups.append(g)

            for sg in g.extended_groups():
                if sg not in groups:
                    groups.append(sg)

        return groups


class Defaults(BaseAttributes):
    __slots__ = ("data", "connection_options")

    def __init__(
        self,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        platform: Optional[str] = None,
        data: Optional[Dict[str, Any]] = None,
        connection_options: Optional[Dict[str, ConnectionOptions]] = None,
    ) -> None:
        self.data = data or {}
        self.connection_options = connection_options or {}
        super().__init__(
            hostname=hostname,
            port=port,
            username=username,
            password=password,
            platform=platform,
        )

    @classmethod
    def schema(self) -> Dict[str, Any]:
        return {
            "data": {"$key": "$value"},
            "connection_options": {"$connection_type": ConnectionOptions.schema()},
            **super().schema(),
        }

    def dict(self) -> Dict[str, Any]:
        return {
            "data": self.data,
            "connection_options": {
                k: v.dict() for k, v in self.connection_options.items()
            },
            **super().dict(),
        }


class Host(InventoryElement):
    __slots__ = ("name", "connections", "defaults")

    def __init__(
        self,
        name: str,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        platform: Optional[str] = None,
        groups: Optional[ParentGroups] = None,
        data: Optional[Dict[str, Any]] = None,
        connection_options: Optional[Dict[str, ConnectionOptions]] = None,
        defaults: Optional[Defaults] = None,
    ) -> None:
        self.name = name
        self.defaults = defaults or Defaults(None, None, None, None, None, None, None)
        self.connections: Dict[str, ConnectionPlugin] = {}
        super().__init__(
            hostname=hostname,
            port=port,
            username=username,
            password=password,
            platform=platform,
            groups=groups,
            data=data,
            connection_options=connection_options,
        )

    def extended_data(self) -> Dict[str, Any]:
        """
        Returns the data associated with the object including inherited data
        """
        processed = []
        result = {}
        for k, v in self.data.items():
            processed.append(k)
            result[k] = v
        for g in self.extended_groups():
            for k, v in g.data.items():
                if k not in processed:
                    processed.append(k)
                    result[k] = v
        for k, v in self.defaults.data.items():
            if k not in processed:
                processed.append(k)
                result[k] = v
        return result

    @classmethod
    def schema(cls) -> Dict[str, Any]:
        return {
            "name": "str",
            "connection_options": {"$connection_type": ConnectionOptions.schema()},
            **super().schema(),
        }

    def dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "connection_options": {
                k: v.dict() for k, v in self.connection_options.items()
            },
            **super().dict(),
        }

    def keys(self) -> KeysView[str]:
        """Returns the keys of the attribute ``data`` and of the parent(s) groups."""
        return self.extended_data().keys()

    def values(self) -> ValuesView[Any]:
        """Returns the values of the attribute ``data`` and of the parent(s) groups."""
        return self.extended_data().values()

    def items(self) -> ItemsView[str, Any]:
        """
        Returns all the data accessible from a device, including
        the one inherited from parent groups
        """
        return self.extended_data().items()

    def has_parent_group(self, group: Union[str, "Group"]) -> bool:
        """Returns whether the object is a child of the :obj:`Group` ``group``"""
        if isinstance(group, str):
            return self._has_parent_group_by_name(group)

        else:
            return self._has_parent_group_by_object(group)

    def _has_parent_group_by_name(self, group: str) -> bool:
        for g in self.groups:
            if g.name == group or g.has_parent_group(group):
                return True
        return False

    def _has_parent_group_by_object(self, group: "Group") -> bool:
        for g in self.groups:
            if g is group or g.has_parent_group(group):
                return True
        return False

    def __getitem__(self, item: str) -> Any:
        try:
            return self.data[item]

        except KeyError:
            for g in self.extended_groups():
                try:
                    r = g.data[item]
                    return r
                except KeyError:
                    continue

            r = self.defaults.data.get(item)
            if r is not None:
                return r

            raise

    def __getattribute__(self, name: str) -> Any:
        if name not in ("hostname", "port", "username", "password", "platform"):
            return object.__getattribute__(self, name)
        v = object.__getattribute__(self, name)
        if v is None:
            for g in self.extended_groups():
                r = object.__getattribute__(g, name)
                if r is not None:
                    return r

            return object.__getattribute__(self.defaults, name)
        else:
            return v

    def __bool__(self) -> bool:
        return bool(self.name)

    def __setitem__(self, item: str, value: Any) -> None:
        self.data[item] = value

    def __len__(self) -> int:
        return len(self.extended_data().keys())

    def __iter__(self) -> Iterator[str]:
        return self.extended_data().__iter__()

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return "{}: {}".format(self.__class__.__name__, self.name or "")

    def get(self, item: str, default: Any = None) -> Any:
        """
        Returns the value ``item`` from the host or hosts group variables.

        Arguments:
            item(``str``): The variable to get
            default(``any``): Return value if item not found
        """
        if hasattr(self, item):
            return getattr(self, item)
        try:
            return self.__getitem__(item)

        except KeyError:
            return default

    def get_connection_parameters(
        self, connection: Optional[str] = None
    ) -> ConnectionOptions:
        if not connection:
            d = ConnectionOptions(
                hostname=self.hostname,
                port=self.port,
                username=self.username,
                password=self.password,
                platform=self.platform,
                extras={},
            )
        else:
            r = self._get_connection_options_recursively(connection)
            if r is not None:
                d = ConnectionOptions(
                    hostname=r.hostname if r.hostname is not None else self.hostname,
                    port=r.port if r.port is not None else self.port,
                    username=r.username if r.username is not None else self.username,
                    password=r.password if r.password is not None else self.password,
                    platform=r.platform if r.platform is not None else self.platform,
                    extras=r.extras if r.extras is not None else {},
                )
            else:
                d = ConnectionOptions(
                    hostname=self.hostname,
                    port=self.port,
                    username=self.username,
                    password=self.password,
                    platform=self.platform,
                    extras={},
                )
        return d

    def _get_connection_options_recursively(
        self, connection: str
    ) -> Optional[ConnectionOptions]:
        p = self.connection_options.get(connection)
        if p is None:
            p = ConnectionOptions(None, None, None, None, None, None)

        for g in self.groups:
            sp = g._get_connection_options_recursively(connection)
            if sp is not None:
                p.hostname = p.hostname if p.hostname is not None else sp.hostname
                p.port = p.port if p.port is not None else sp.port
                p.username = p.username if p.username is not None else sp.username
                p.password = p.password if p.password is not None else sp.password
                p.platform = p.platform if p.platform is not None else sp.platform
                p.extras = p.extras if p.extras is not None else sp.extras

        sp = self.defaults.connection_options.get(connection, None)
        if sp is not None:
            p.hostname = p.hostname if p.hostname is not None else sp.hostname
            p.port = p.port if p.port is not None else sp.port
            p.username = p.username if p.username is not None else sp.username
            p.password = p.password if p.password is not None else sp.password
            p.platform = p.platform if p.platform is not None else sp.platform
            p.extras = p.extras if p.extras is not None else sp.extras
        return p

    def get_connection(self, connection: str, configuration: Config) -> Any:
        """
        The function of this method is twofold:

            1. If an existing connection is already established for the given type return it
            2. If none exists, establish a new connection of that type with default parameters
               and return it

        Raises:
            AttributeError: if it's unknown how to establish a connection for the given type

        Arguments:
            connection: Name of the connection, for instance, netmiko, paramiko, napalm...

        Returns:
            An already established connection
        """
        if connection not in self.connections:
            conn = self.get_connection_parameters(connection)
            self.open_connection(
                connection=connection,
                configuration=configuration,
                hostname=conn.hostname,
                port=conn.port,
                username=conn.username,
                password=conn.password,
                platform=conn.platform,
                extras=conn.extras,
            )
        return self.connections[connection].connection

    def open_connection(
        self,
        connection: str,
        configuration: Config,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        port: Optional[int] = None,
        platform: Optional[str] = None,
        extras: Optional[Dict[str, Any]] = None,
        default_to_host_attributes: bool = True,
    ) -> ConnectionPlugin:
        """
        Open a new connection.

        If ``default_to_host_attributes`` is set to ``True`` arguments will default to host
        attributes if not specified.

        Raises:
            AttributeError: if it's unknown how to establish a connection for the given type

        Returns:
            An already established connection
        """
        conn_name = connection
        existing_conn = self.connections.get(conn_name)
        if existing_conn is not None:
            raise ConnectionAlreadyOpen(conn_name)

        plugin = ConnectionPluginRegister.get_plugin(conn_name)
        conn_obj = plugin()
        if default_to_host_attributes:
            conn_params = self.get_connection_parameters(conn_name)
            hostname = hostname if hostname is not None else conn_params.hostname
            username = username if username is not None else conn_params.username
            password = password if password is not None else conn_params.password
            port = port if port is not None else conn_params.port
            platform = platform if platform is not None else conn_params.platform
            extras = extras if extras is not None else conn_params.extras

        conn_obj.open(
            hostname=hostname,
            username=username,
            password=password,
            port=port,
            platform=platform,
            extras=extras,
            configuration=configuration,
        )
        self.connections[conn_name] = conn_obj
        return conn_obj

    def close_connection(self, connection: str) -> None:
        """ Close the connection"""
        conn_name = connection
        if conn_name not in self.connections:
            raise ConnectionNotOpen(conn_name)

        conn_obj = self.connections.pop(conn_name)
        if conn_obj is not None:
            conn_obj.close()

    def close_connections(self) -> None:
        # Decouple deleting dictionary elements from iterating over connections dict
        existing_conns = list(self.connections.keys())
        for connection in existing_conns:
            self.close_connection(connection)


class Group(Host):
    pass


class Hosts(Dict[str, Host]):
    pass


class Groups(Dict[str, Group]):
    pass


TransformFunction = Callable[[Arg(Host), KwArg(Any)], None]
FilterObj = Callable[[Arg(Host), KwArg(Any)], bool]


class Inventory(object):
    __slots__ = ("hosts", "groups", "defaults")

    def __init__(
        self,
        hosts: Hosts,
        groups: Optional[Groups] = None,
        defaults: Optional[Defaults] = None,
        transform_function: TransformFunction = None,
        transform_function_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.hosts = hosts
        self.groups = groups or Groups()
        self.defaults = defaults or Defaults(None, None, None, None, None, None, None)

    def filter(
        self, filter_obj: FilterObj = None, filter_func: FilterObj = None, **kwargs: Any
    ) -> "Inventory":
        filter_func = filter_obj or filter_func
        if filter_func:
            filtered = Hosts(
                {n: h for n, h in self.hosts.items() if filter_func(h, **kwargs)}
            )
        else:
            filtered = Hosts(
                {
                    n: h
                    for n, h in self.hosts.items()
                    if all(h.get(k) == v for k, v in kwargs.items())
                }
            )
        return Inventory(hosts=filtered, groups=self.groups, defaults=self.defaults)

    def __len__(self) -> int:
        return self.hosts.__len__()

    def children_of_group(self, group: Union[str, Group]) -> Set[Host]:
        """
        Returns set of hosts that belongs to a group including those that belong
        indirectly via inheritance
        """
        hosts: Set[Host] = set()
        for host in self.hosts.values():
            if host.has_parent_group(group):
                hosts.add(host)
        return hosts

    @classmethod
    def schema(cls) -> Dict[str, Any]:
        """
        Return serialized dictionary of inventory
        """
        return {
            "hosts": {"$name": Host.schema()},
            "groups": {"$group": Group.schema()},
            "defaults": Defaults.schema(),
        }

    def dict(self) -> Dict[str, Any]:
        """
        Return serialized dictionary of inventory
        """
        return {
            "hosts": {n: h.dict() for n, h in self.hosts.items()},
            "groups": {n: g.dict() for n, g in self.groups.items()},
            "defaults": self.defaults.dict(),
        }



from typing import Any, Dict, Type, Optional

from nornir.core.configuration import Config
from nornir.core.plugins.register import PluginRegister

from typing_extensions import Protocol


CONNECTIONS_PLUGIN_PATH = "nornir.plugins.connections"


class ConnectionPlugin(Protocol):
    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:
        """
        Connect to the device and populate the attribute :attr:`connection` with
        the underlying connection
        """
        pass

    def close(self) -> None:
        """Close the connection with the device"""
        pass

    @property
    def connection(self) -> Any:
        """
        Established connection
        """
        pass


ConnectionPluginRegister: PluginRegister[Type[ConnectionPlugin]] = PluginRegister(
    CONNECTIONS_PLUGIN_PATH
)



from typing import Any

from nornir.core import Nornir
from nornir.core.configuration import Config
from nornir.core.inventory import Inventory
from nornir.core.plugins.inventory import (
    InventoryPluginRegister,
    TransformFunctionRegister,
)
from nornir.core.plugins.connections import ConnectionPluginRegister
from nornir.core.plugins.runners import RunnerPlugin, RunnersPluginRegister
from nornir.core.state import GlobalState


def load_inventory(config: Config,) -> Inventory:
    InventoryPluginRegister.auto_register()
    inventory_plugin = InventoryPluginRegister.get_plugin(config.inventory.plugin)
    inv = inventory_plugin(**config.inventory.options).load()

    if config.inventory.transform_function:
        TransformFunctionRegister.auto_register()
        transform_function = TransformFunctionRegister.get_plugin(
            config.inventory.transform_function
        )
        for h in inv.hosts.values():
            transform_function(h, **(config.inventory.transform_function_options or {}))

    return inv


def load_runner(config: Config,) -> RunnerPlugin:
    RunnersPluginRegister.auto_register()
    runner_plugin = RunnersPluginRegister.get_plugin(config.runner.plugin)
    runner = runner_plugin(**config.runner.options)
    return runner


def InitNornir(config_file: str = "", dry_run: bool = False, **kwargs: Any,) -> Nornir:
    """
    Arguments:
        config_file(str): Path to the configuration file (optional)
        dry_run(bool): Whether to simulate changes or not
        configure_logging: Whether to configure logging or not. This argument is being
            deprecated. Please use logging.enabled parameter in the configuration
            instead.
        **kwargs: Extra information to pass to the
            :obj:`nornir.core.configuration.Config` object

    Returns:
        :obj:`nornir.core.Nornir`: fully instantiated and configured
    """
    ConnectionPluginRegister.auto_register()

    if config_file:
        config = Config.from_file(config_file, **kwargs)
    else:
        config = Config.from_dict(**kwargs)

    data = GlobalState(dry_run=dry_run)

    config.logging.configure()

    return Nornir(
        inventory=load_inventory(config),
        runner=load_runner(config),
        config=config,
        data=data,
    )



import logging
import traceback
from typing import Any, Callable, Dict, List, Optional, TYPE_CHECKING, Union, cast

from nornir.core.exceptions import NornirExecutionError
from nornir.core.exceptions import NornirSubTaskError

if TYPE_CHECKING:
    from nornir.core import Nornir
    from nornir.core.inventory import Host
    from nornir.core.processor import Processors


logger = logging.getLogger(__name__)
DEFAULT_SEVERITY_LEVEL = logging.INFO


class Task(object):
    """
    A task is basically a wrapper around a function that has to be run against multiple devices.
    You won't probably have to deal with this class yourself as
    :meth:`nornir.core.Nornir.run` will create it automatically.

    Arguments:
        task (callable): function or callable we will be calling
        name (``string``): name of task, defaults to ``task.__name__``
        severity_level (logging.LEVEL): Severity level associated to the task
        **kwargs: Parameters that will be passed to the ``task``

    Attributes:
        task (callable): function or callable we will be calling
        name (``string``): name of task, defaults to ``task.__name__``
        params: Parameters that will be passed to the ``task``.
        self.results (:obj:`nornir.core.task.MultiResult`): Intermediate results
        host (:obj:`nornir.core.inventory.Host`): Host we are operating with. Populated right
          before calling the ``task``
        nornir(:obj:`nornir.core.Nornir`): Populated right before calling
          the ``task``
        severity_level (logging.LEVEL): Severity level associated to the task
    """

    def __init__(
        self,
        task: Callable[..., Any],
        nornir: "Nornir",
        global_dry_run: bool,
        processors: "Processors",
        name: str = None,
        severity_level: int = DEFAULT_SEVERITY_LEVEL,
        parent_task: Optional["Task"] = None,
        **kwargs: str
    ):
        self.task = task
        self.nornir = nornir
        self.name = name or task.__name__
        self.global_dry_run = global_dry_run
        self.parent_task = parent_task
        self.params = kwargs
        self.results = MultiResult(self.name)
        self.severity_level = severity_level
        self.processors = processors

    def copy(self) -> "Task":
        return Task(
            self.task,
            self.nornir,
            self.global_dry_run,
            self.processors,
            self.name,
            self.severity_level,
            self.parent_task,
            **self.params
        )

    def __repr__(self) -> str:
        return self.name

    def start(self, host: "Host") -> "MultiResult":
        """
        Run the task for the given host.

        Arguments:
            host (:obj:`nornir.core.inventory.Host`): Host we are operating with. Populated right
              before calling the ``task``
            nornir(:obj:`nornir.core.Nornir`): Populated right before calling
              the ``task``

        Returns:
            host (:obj:`nornir.core.task.MultiResult`): Results of the task and its subtasks
        """
        self.host = host

        if self.parent_task is not None:
            self.processors.subtask_instance_started(self, host)
        else:
            self.processors.task_instance_started(self, host)
        try:
            logger.debug("Host %r: running task %r", self.host.name, self.name)
            r = self.task(self, **self.params)
            if not isinstance(r, Result):
                r = Result(host=host, result=r)

        except NornirSubTaskError as e:
            tb = traceback.format_exc()
            logger.error(
                "Host %r: task %r failed with traceback:\n%s",
                self.host.name,
                self.name,
                tb,
            )
            r = Result(host, exception=e, result=str(e), failed=True)

        except Exception as e:
            tb = traceback.format_exc()
            logger.error(
                "Host %r: task %r failed with traceback:\n%s",
                self.host.name,
                self.name,
                tb,
            )
            r = Result(host, exception=e, result=tb, failed=True)

        r.name = self.name

        if r.severity_level == DEFAULT_SEVERITY_LEVEL:
            if r.failed:
                r.severity_level = logging.ERROR
            else:
                r.severity_level = self.severity_level

        self.results.insert(0, r)

        if self.parent_task is not None:
            self.processors.subtask_instance_completed(self, host, self.results)
        else:
            self.processors.task_instance_completed(self, host, self.results)
        return self.results

    def run(self, task: Callable[..., Any], **kwargs: Any) -> "MultiResult":
        """
        This is a utility method to call a task from within a task. For instance:

            def grouped_tasks(task):
                task.run(my_first_task)
                task.run(my_second_task)

            nornir.run(grouped_tasks)

        This method will ensure the subtask is run only for the host in the current thread.
        """
        if not self.host:
            msg = (
                "You have to call this after setting host and nornir attributes. ",
                "You probably called this from outside a nested task",
            )
            raise Exception(msg)

        if "severity_level" not in kwargs:
            kwargs["severity_level"] = self.severity_level

        run_task = Task(
            task,
            self.nornir,
            global_dry_run=self.global_dry_run,
            processors=self.processors,
            parent_task=self,
            **kwargs
        )
        r = run_task.start(self.host)
        self.results.append(r[0] if len(r) == 1 else cast("Result", r))

        if r.failed:
            # Without this we will keep running the grouped task
            raise NornirSubTaskError(task=run_task, result=r)

        return r

    def is_dry_run(self, override: Optional[bool] = None) -> bool:
        """
        Returns whether current task is a dry_run or not.
        """
        return override if override is not None else self.global_dry_run


class Result(object):
    """
    Result of running individual tasks.

    Arguments:
        changed (bool): ``True`` if the task is changing the system
        diff (obj): Diff between state of the system before/after running this task
        result (obj): Result of the task execution, see task's documentation for details
        host (:obj:`nornir.core.inventory.Host`): Reference to the host that lead to this result
        failed (bool): Whether the execution failed or not
        severity_level (logging.LEVEL): Severity level associated to the result of the excecution
        exception (Exception): uncaught exception thrown during the exection of the task (if any)

    Attributes:
        changed (bool): ``True`` if the task is changing the system
        diff (obj): Diff between state of the system before/after running this task
        result (obj): Result of the task execution, see task's documentation for details
        host (:obj:`nornir.core.inventory.Host`): Reference to the host that lead ot this result
        failed (bool): Whether the execution failed or not
        severity_level (logging.LEVEL): Severity level associated to the result of the excecution
        exception (Exception): uncaught exception thrown during the exection of the task (if any)
    """

    def __init__(
        self,
        host: Union["Host", None],
        result: Any = None,
        changed: bool = False,
        diff: str = "",
        failed: bool = False,
        exception: Optional[BaseException] = None,
        severity_level: int = DEFAULT_SEVERITY_LEVEL,
        **kwargs: Any
    ):
        self.result = result
        self.host = host
        self.changed = changed
        self.diff = diff
        self.failed = failed
        self.exception = exception
        self.name = None
        self.severity_level = severity_level

        self.stdout: Optional[str] = None
        self.stderr: Optional[str] = None

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self) -> str:
        return '{}: "{}"'.format(self.__class__.__name__, self.name)

    def __str__(self) -> str:
        if self.exception:
            return str(self.exception)

        else:
            return str(self.result)


class MultiResult(List[Result]):
    """
    It is basically is a list-like object that gives you access to the results of all subtasks for
    a particular device/task.
    """

    def __init__(self, name: str):
        self.name = name

    def __getattr__(self, name: str) -> Any:
        return getattr(self[0], name)

    def __repr__(self) -> str:
        return "{}: {}".format(self.__class__.__name__, super().__repr__())

    @property
    def failed(self) -> bool:
        """If ``True`` at least a task failed."""
        return any([h.failed for h in self])

    @property
    def changed(self) -> bool:
        """If ``True`` at least a task changed the system."""
        return any([h.changed for h in self])

    def raise_on_error(self) -> None:
        """
        Raises:
            :obj:`nornir.core.exceptions.NornirExecutionError`: When at least a task failed
        """
        if self.failed:
            raise NornirExecutionError(self)


class AggregatedResult(Dict[str, MultiResult]):
    """
    It basically is a dict-like object that aggregates the results for all devices.
    You can access each individual result by doing ``my_aggr_result["hostname_of_device"]``.
    """

    def __init__(self, name: str, **kwargs: MultiResult):
        self.name = name
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        return "{} ({}): {}".format(
            self.__class__.__name__, self.name, super().__repr__()
        )

    @property
    def failed(self) -> bool:
        """If ``True`` at least a host failed."""
        return any([h.failed for h in self.values()])

    @property
    def failed_hosts(self) -> Dict[str, "MultiResult"]:
        """Hosts that failed during the execution of the task."""
        return {h: r for h, r in self.items() if r.failed}

    def raise_on_error(self) -> None:
        """
        Raises:
            :obj:`nornir.core.exceptions.NornirExecutionError`: When at least a task failed
        """
        if self.failed:
            raise NornirExecutionError(self)



import ast
import logging
import logging.handlers
import os
import sys
import warnings
from pathlib import Path
from typing import Any, Dict, Optional, Type, TYPE_CHECKING, List, TypeVar

from nornir.core.exceptions import ConflictingConfigurationWarning

import ruamel.yaml

if TYPE_CHECKING:
    from nornir.core.deserializer.inventory import Inventory  # noqa


DEFAULT_SSH_CONFIG = str(Path("~/.ssh/config").expanduser())

T = TypeVar("T")


class Parameter:
    def __init__(
        self,
        envvar: str,
        typ: Optional[Type[T]] = None,
        help: str = "",
        default: Optional[T] = None,
    ) -> None:
        if typ is not None:
            self.type: Type[T] = typ
        elif default is not None:
            self.type = default.__class__
        else:
            raise TypeError("either typ or default needs to be specified")
        self.envvar = envvar
        self.help = help
        self.default = default or self.type()

    def resolve(self, value: Optional[T]) -> T:
        v: Optional[Any] = value
        if value is None:
            t = os.environ.get(self.envvar)
            if self.type is bool and t:
                v = t in ["true", "True", "1", "yes"]
            elif self.type is str and t:
                v = t
            elif t:
                v = ast.literal_eval(t) if t is not None else None

        if v is None:
            v = self.default
        return v


class SSHConfig(object):
    __slots__ = ("config_file",)

    class Parameters:
        config_file = Parameter(
            default=DEFAULT_SSH_CONFIG, envvar="NORNIR_SSH_CONFIG_FILE"
        )

    def __init__(self, config_file: Optional[str] = None) -> None:
        self.config_file = self.Parameters.config_file.resolve(config_file)

    def dict(self) -> Dict[str, Any]:
        return {"config_file": self.config_file}


class InventoryConfig(object):
    __slots__ = "plugin", "options", "transform_function", "transform_function_options"

    class Parameters:
        plugin = Parameter(
            typ=str, default="SimpleInventory", envvar="NORNIR_INVENTORY_PLUGIN"
        )
        options = Parameter(default={}, envvar="NORNIR_INVENTORY_OPTIONS")
        transform_function = Parameter(
            typ=str, envvar="NORNIR_INVENTORY_TRANSFORM_FUNCTION"
        )
        transform_function_options = Parameter(
            default={}, envvar="NORNIR_INVENTORY_TRANSFORM_FUNCTION_OPTIONS"
        )

    def __init__(
        self,
        plugin: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        transform_function: Optional[str] = None,
        transform_function_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.plugin = self.Parameters.plugin.resolve(plugin)
        self.options = self.Parameters.options.resolve(options) or {}
        self.transform_function = self.Parameters.transform_function.resolve(
            transform_function
        )
        self.transform_function_options = self.Parameters.transform_function_options.resolve(
            transform_function_options
        )

    def dict(self) -> Dict[str, Any]:
        return {
            "plugin": self.plugin,
            "options": self.options,
            "transform_function": self.transform_function,
            "transform_function_options": self.transform_function_options,
        }


class LoggingConfig(object):
    __slots__ = "enabled", "level", "log_file", "format", "to_console", "loggers"

    class Parameters:
        enabled = Parameter(default=True, envvar="NORNIR_LOGGING_ENABLED")
        level = Parameter(default="INFO", envvar="NORNIR_LOGGING_LEVEL")
        log_file = Parameter(default="nornir.log", envvar="NORNIR_LOGGING_LOG_FILE")
        format = Parameter(
            default="%(asctime)s - %(name)12s - %(levelname)8s - %(funcName)10s() - %(message)s",
            envvar="NORNIR_LOGGING_FORMAT",
        )
        to_console = Parameter(default=False, envvar="NORNIR_LOGGING_TO_CONSOLE")
        loggers = Parameter(default=["nornir"], envvar="NORNIR_LOGGING_LOGGERS")

    def __init__(
        self,
        enabled: Optional[bool] = None,
        level: Optional[str] = None,
        log_file: Optional[str] = None,
        format: Optional[str] = None,
        to_console: Optional[bool] = None,
        loggers: Optional[List[str]] = None,
    ) -> None:
        self.enabled = self.Parameters.enabled.resolve(enabled)
        self.level = self.Parameters.level.resolve(level)
        self.log_file = self.Parameters.log_file.resolve(log_file)
        self.format = self.Parameters.format.resolve(format)
        self.to_console = self.Parameters.to_console.resolve(to_console)
        self.loggers = self.Parameters.loggers.resolve(loggers)

    def dict(self) -> Dict[str, Any]:
        return {
            "enabled": self.enabled,
            "level": self.level,
            "log_file": self.log_file,
            "format": self.format,
            "to_console": self.to_console,
            "loggers": self.loggers,
        }

    def configure(self) -> None:
        if not self.enabled:
            return

        root_logger = logging.getLogger()
        if root_logger.hasHandlers() or root_logger.level != logging.WARNING:
            msg = (
                "Native Python logging configuration has been detected, but Nornir "
                "logging is enabled too. "
                "This can lead to unexpected logging results. "
                "Please set logging.enabled config to False "
                "to disable automatic Nornir logging configuration. Refer to "
                "https://nornir.readthedocs.io/en/stable/configuration/index.html#logging"  # noqa
            )
            warnings.warn(msg, ConflictingConfigurationWarning)

        formatter = logging.Formatter(self.format)
        # log INFO and DEBUG to stdout
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(formatter)
        stdout_handler.setLevel(logging.DEBUG)
        stdout_handler.addFilter(lambda record: record.levelno <= logging.INFO)
        # log WARNING, ERROR and CRITICAL to stderr
        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.setFormatter(formatter)
        stderr_handler.setLevel(logging.WARNING)

        for logger_name in self.loggers:
            logger_ = logging.getLogger(logger_name)
            logger_.propagate = False
            logger_.setLevel(self.level)
            if logger_.hasHandlers():
                # Don't add handlers if some handlers are already attached to the logger
                # This is crucial to avoid duplicate handlers
                # Alternative would be to clear all handlers and reconfigure them
                # with Nornir
                # There are several situations this branch can be executed:
                # multiple calls to InitNornir,
                # logging.config.dictConfig configuring 'nornir' logger, etc.
                # The warning is not emitted in this scenario
                continue
            if self.log_file:
                handler = logging.handlers.RotatingFileHandler(
                    str(Path(self.log_file)), maxBytes=1024 * 1024 * 10, backupCount=20
                )
                handler.setFormatter(formatter)
                logger_.addHandler(handler)

            if self.to_console:
                logger_.addHandler(stdout_handler)
                logger_.addHandler(stderr_handler)


class RunnerConfig(object):
    __slots__ = ("plugin", "options")

    class Parameters:
        plugin = Parameter(default="threaded", envvar="NORNIR_RUNNER_PLUGIN")
        options = Parameter(default={}, envvar="NORNIR_RUNNER_OPTIONS")

    def __init__(
        self, plugin: Optional[str] = None, options: Optional[Dict[str, Any]] = None
    ) -> None:
        self.plugin = self.Parameters.plugin.resolve(plugin)
        self.options = self.Parameters.options.resolve(options)

    def dict(self) -> Dict[str, Any]:
        return {
            "plugin": self.plugin,
            "options": self.options,
        }


class CoreConfig(object):
    __slots__ = "raise_on_error"

    class Parameters:
        raise_on_error = Parameter(default=False, envvar="NORNIR_CORE_RAISE_ON_ERROR")

    def __init__(self, raise_on_error: Optional[bool] = None) -> None:
        self.raise_on_error = self.Parameters.raise_on_error.resolve(raise_on_error)

    def dict(self) -> Dict[str, Any]:
        return {
            "raise_on_error": self.raise_on_error,
        }


class Config(object):
    __slots__ = (
        "core",
        "runner",
        "ssh",
        "inventory",
        "jinja2",
        "logging",
        "user_defined",
    )

    def __init__(
        self,
        inventory: Optional[InventoryConfig] = None,
        ssh: Optional[SSHConfig] = None,
        logging: Optional[LoggingConfig] = None,
        core: Optional[CoreConfig] = None,
        runner: Optional[RunnerConfig] = None,
        user_defined: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.inventory = inventory or InventoryConfig()
        self.ssh = ssh or SSHConfig()
        self.logging = logging or LoggingConfig()
        self.core = core or CoreConfig()
        self.runner = runner or RunnerConfig()
        self.user_defined = user_defined or {}

    @classmethod
    def from_dict(
        cls,
        inventory: Dict[str, Any] = None,
        ssh: Optional[Dict[str, Any]] = None,
        logging: Optional[Dict[str, Any]] = None,
        core: Optional[Dict[str, Any]] = None,
        runner: Optional[Dict[str, Any]] = None,
        user_defined: Optional[Dict[str, Any]] = None,
    ) -> "Config":
        return cls(
            inventory=InventoryConfig(**inventory or {}),
            ssh=SSHConfig(**ssh or {}),
            logging=LoggingConfig(**logging or {}),
            core=CoreConfig(**core or {}),
            runner=RunnerConfig(**runner or {}),
            user_defined=user_defined or {},
        )

    @classmethod
    def from_file(
        cls,
        config_file: str,
        inventory: Optional[Dict[str, Any]] = None,
        ssh: Optional[Dict[str, Any]] = None,
        logging: Optional[Dict[str, Any]] = None,
        core: Optional[Dict[str, Any]] = None,
        runner: Optional[Dict[str, Any]] = None,
        user_defined: Optional[Dict[str, Any]] = None,
    ) -> "Config":
        inventory = inventory or {}
        ssh = ssh or {}
        logging = logging or {}
        core = core or {}
        runner = runner or {}
        user_defined = user_defined or {}
        with open(config_file, "r") as f:
            yml = ruamel.yaml.YAML(typ="safe")
            data = yml.load(f)
        return cls(
            inventory=InventoryConfig(**{**data.get("inventory", {}), **inventory}),
            ssh=SSHConfig(**{**data.get("ssh", {}), **ssh}),
            logging=LoggingConfig(**{**data.get("logging", {}), **logging}),
            core=CoreConfig(**{**data.get("core", {}), **core}),
            runner=RunnerConfig(**{**data.get("runner", {}), **runner}),
            user_defined={**data.get("user_defined", {}), **user_defined},
        )

    def dict(self) -> Dict[str, Any]:
        return {
            "inventory": self.inventory.dict(),
            "ssh": self.ssh.dict(),
            "logging": self.logging.dict(),
            "core": self.core.dict(),
            "runner": self.runner.dict(),
            "user_defined": self.user_defined,
        }



from typing import Any, Dict, Optional

from jinja2 import Environment, FileSystemLoader, StrictUndefined


def render_from_file(
    path: str,
    template: str,
    jinja_filters: Optional[Dict[str, Any]] = None,
    **kwargs: Any
) -> str:
    jinja_filters = jinja_filters or {}
    env = Environment(
        loader=FileSystemLoader(path), undefined=StrictUndefined, trim_blocks=True
    )
    env.filters.update(jinja_filters)
    t = env.get_template(template)
    return t.render(**kwargs)


def render_from_string(
    template: str, jinja_filters: Optional[Dict[str, Any]] = None, **kwargs: Any
) -> str:
    jinja_filters = jinja_filters or {}
    env = Environment(undefined=StrictUndefined, trim_blocks=True)
    env.filters.update(jinja_filters)
    t = env.from_string(template)
    return t.render(**kwargs)



import pkg_resources
from typing import Dict, TypeVar, Generic


from nornir.core.exceptions import (
    PluginAlreadyRegistered,
    PluginNotRegistered,
)

T = TypeVar("T")


class PluginRegister(Generic[T]):
    available: Dict[str, T] = {}

    def __init__(self, entry_point: str) -> None:
        self._entry_point = entry_point

    def auto_register(self) -> None:
        discovered_plugins: Dict[str, T] = {
            entry_point.name: entry_point.load()
            for entry_point in pkg_resources.iter_entry_points(self._entry_point)
        }
        for k, v in discovered_plugins.items():
            self.register(k, v)

    def register(self, name: str, plugin: T) -> None:
        """Registers a plugin with a specified name

        Args:
            name: name of the connection plugin to register
            plugin: plugin class

        Raises:
            :obj:`nornir.core.exceptions.PluginAlreadyRegistered` if
                another plugin with the specified name was already registered
        """
        existing_plugin = self.available.get(name)
        if existing_plugin is None:
            self.available[name] = plugin
        elif existing_plugin != plugin:
            raise PluginAlreadyRegistered(
                f"plugin {plugin} can't be registered as "
                f"{name!r} because plugin {existing_plugin} "
                f"was already registered under this name"
            )

    def deregister(self, name: str) -> None:
        """Deregisters a registered plugin by its name

        Args:
            name: name of the plugin to deregister

        Raises:
            :obj:`nornir.core.exceptions.PluginNotRegistered`
        """
        if name not in self.available:
            raise PluginNotRegistered(f"plugin {name!r} is not registered")
        self.available.pop(name)

    def deregister_all(self) -> None:
        """Deregisters all registered plugins"""
        self.available = {}

    def get_plugin(self, name: str) -> T:
        """Fetches the plugin by name if already registered

        Args:
            name: name of the plugin

        Raises:
            :obj:`nornir.core.exceptions.PluginNotRegistered`
        """
        if name not in self.available:
            raise PluginNotRegistered(f"plugin {name!r} is not registered")
        return self.available[name]



import logging
import logging.config
from typing import List, Optional, TYPE_CHECKING

from nornir.core.configuration import Config
from nornir.core.inventory import Inventory
from nornir.core.plugins.runners import RunnerPlugin
from nornir.core.processor import Processor, Processors
from nornir.core.state import GlobalState
from nornir.core.task import Task

if TYPE_CHECKING:
    from nornir.core.inventory import Host  # noqa: W0611

logger = logging.getLogger(__name__)


class Nornir(object):
    """
    This is the main object to work with. It contains the inventory and it serves
    as task dispatcher.

    Arguments:
        inventory (:obj:`nornir.core.inventory.Inventory`): Inventory to work with
        data(GlobalState): shared data amongst different iterations of nornir
        dry_run(``bool``): Whether if we are testing the changes or not
        config (:obj:`nornir.core.configuration.Config`): Configuration object

    Attributes:
        inventory (:obj:`nornir.core.inventory.Inventory`): Inventory to work with
        data(:obj:`nornir.core.GlobalState`): shared data amongst different iterations of nornir
        dry_run(``bool``): Whether if we are testing the changes or not
        config (:obj:`nornir.core.configuration.Config`): Configuration parameters
    """

    def __init__(
        self,
        inventory: Inventory,
        config: Config = None,
        data: GlobalState = None,
        processors: Optional[Processors] = None,
        runner: Optional[RunnerPlugin] = None,
    ) -> None:
        self.data = data if data is not None else GlobalState()
        self.inventory = inventory
        self.config = config or Config()
        self.processors = processors or Processors()
        self.runner = runner

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_connections(on_good=True, on_failed=True)

    def with_processors(self, processors: List[Processor]) -> "Nornir":
        """
        Given a list of Processor objects return a copy of the nornir object with the processors
        assigned to the copy. The orinal object is left unmodified.
        """
        return Nornir(**{**self.__dict__, **{"processors": Processors(processors)}})

    def with_runner(self, runner: RunnerPlugin) -> "Nornir":
        """
        Given a runner return a copy of the nornir object with the runner
        assigned to the copy. The orinal object is left unmodified.
        """
        return Nornir(**{**self.__dict__, **{"runner": runner}})

    def filter(self, *args, **kwargs):
        """
        See :py:meth:`nornir.core.inventory.Inventory.filter`

        Returns:
            :obj:`Nornir`: A new object with same configuration as ``self`` but filtered inventory.
        """
        b = Nornir(**self.__dict__)
        b.inventory = self.inventory.filter(*args, **kwargs)
        return b

    def run(
        self,
        task,
        raise_on_error=None,
        on_good=True,
        on_failed=False,
        name: Optional[str] = None,
        **kwargs,
    ):
        """
        Run task over all the hosts in the inventory.

        Arguments:
            task (``callable``): function or callable that will be run against each device in
              the inventory
            raise_on_error (``bool``): Override raise_on_error behavior
            on_good(``bool``): Whether to run or not this task on hosts marked as good
            on_failed(``bool``): Whether to run or not this task on hosts marked as failed
            **kwargs: additional argument to pass to ``task`` when calling it

        Raises:
            :obj:`nornir.core.exceptions.NornirExecutionError`: if at least a task fails
              and self.config.core.raise_on_error is set to ``True``

        Returns:
            :obj:`nornir.core.task.AggregatedResult`: results of each execution
        """
        task = Task(
            task,
            self,
            global_dry_run=self.data.dry_run,
            name=name,
            processors=self.processors,
            **kwargs,
        )
        self.processors.task_started(task)

        run_on = []
        if on_good:
            for name, host in self.inventory.hosts.items():
                if name not in self.data.failed_hosts:
                    run_on.append(host)
        if on_failed:
            for name, host in self.inventory.hosts.items():
                if name in self.data.failed_hosts:
                    run_on.append(host)

        num_hosts = len(self.inventory.hosts)
        if num_hosts:
            logger.info(
                "Running task %r with args %s on %d hosts",
                task.name,
                kwargs,
                num_hosts,
            )
        else:
            logger.warning("Task %r has not been run – 0 hosts selected", task.name)

        result = self.runner.run(task, run_on)

        raise_on_error = (
            raise_on_error
            if raise_on_error is not None
            else self.config.core.raise_on_error
        )  # noqa
        if raise_on_error:
            result.raise_on_error()
        else:
            self.data.failed_hosts.update(result.failed_hosts.keys())

        self.processors.task_completed(task, result)

        return result

    def dict(self):
        """ Return a dictionary representing the object. """
        return {"data": self.data.dict(), "inventory": self.inventory.dict()}

    def close_connections(self, on_good=True, on_failed=False):
        def close_connections_task(task):
            task.host.close_connections()

        self.run(task=close_connections_task, on_good=on_good, on_failed=on_failed)

    @classmethod
    def get_validators(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, cls):
            raise ValueError(f"Nornir: Nornir expected not {type(v)}")
        return v

    @property
    def state(self):
        return GlobalState



from typing import Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from nornir.core.connection import Connection
    from nornir.core.result import AggregatedResult, MultiResult, Result  # noqa
    from nornir.core.tasks import Task


class ConnectionException(Exception):
    """
    Superclass for all the Connection* Exceptions
    """

    def __init__(self, connection: "Connection") -> None:
        self.connection = connection


class ConnectionAlreadyOpen(ConnectionException):
    """
    Raised when opening an already opened connection
    """

    pass


class ConnectionNotOpen(ConnectionException):
    """
    Raised when trying to close a connection that isn't open
    """

    pass


class PluginAlreadyRegistered(Exception):
    """
    Raised when trying to register an already registered plugin
    """

    pass


class PluginNotRegistered(Exception):
    """
    Raised when trying to access a plugin that is not registered
    """

    pass


class NornirExecutionError(Exception):
    """
    Raised by nornir when any of the tasks managed by :meth:`nornir.core.Nornir.run`
    when any of the tasks fail.
    """

    def __init__(self, result: "AggregatedResult") -> None:
        self.result = result

    @property
    def failed_hosts(self) -> Dict[str, "MultiResult"]:
        """
        Hosts that failed to complete the task
        """
        return {k: v for k, v in self.result.items() if v.failed}

    def __str__(self) -> str:
        text = "\n"
        for k, r in self.result.items():
            text += "{}\n".format("#" * 40)
            if r.failed:
                text += "# {} (failed)\n".format(k)
            else:
                text += "# {} (succeeded)\n".format(k)
            text += "{}\n".format("#" * 40)
            for sub_r in r:
                text += "**** {}\n".format(sub_r.name)
                text += "{}\n".format(sub_r)
        return text


class NornirSubTaskError(Exception):
    """
    Raised by nornir when a sub task managed by :meth:`nornir.core.Task.run` has failed
    """

    def __init__(self, task: "Task", result: "Result"):
        self.task = task
        self.result = result

    def __str__(self) -> str:
        return "Subtask: {} (failed)\n".format(self.task)


class NornirNoValidInventoryError(Exception):
    """
    Raised by nornir when :meth:`nornir.plugins.inventory.parse` fails to load any valid inventory
    """

    pass


class ConflictingConfigurationWarning(UserWarning):
    pass



from typing import Any, Dict


def merge_two_dicts(x: Dict[Any, Any], y: Dict[Any, Any]) -> Dict[Any, Any]:
    try:
        z = x.copy()
    except AttributeError:
        z = dict(x)
    z.update(y)
    return z



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nornir',
 'nornir.core',
 'nornir.core.helpers',
 'nornir.core.plugins',
 'nornir.plugins',
 'nornir.plugins.connections',
 'nornir.plugins.functions',
 'nornir.plugins.inventory',
 'nornir.plugins.processors',
 'nornir.plugins.runners',
 'nornir.plugins.tasks']

package_data = \
{'': ['*']}

install_requires = \
['mypy_extensions>=0.4.1,<0.5.0',
 'ruamel.yaml>=0.16,<0.17',
 'typing_extensions>=3.7,<4.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8'],
 'docs': ['sphinx>=1,<2',
          'sphinx_rtd_theme>=0.4,<0.5',
          'sphinxcontrib-napoleon>=0.7,<0.8',
          'jupyter>=1,<2',
          'nbsphinx>=0.5,<0.6',
          'pygments>=2,<3',
          'sphinx-issues>=1.2,<2.0']}

entry_points = \
{'nornir.plugins.inventory': ['SimpleInventory = '
                              'nornir.plugins.inventory.simple:SimpleInventory'],
 'nornir.plugins.runners': ['serial = nornir.plugins.runners:SerialRunner',
                            'threaded = nornir.plugins.runners:ThreadedRunner']}

setup_kwargs = {
    'name': 'nornir',
    'version': '3.1.0',
    'description': 'Pluggable multi-threaded framework with inventory management to help operate collections of devices',
    'long_description': '![Build Status](https://github.com/nornir-automation/nornir/workflows/test%20nornir/badge.svg) [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black) [![Coverage Status](https://coveralls.io/repos/github/nornir-automation/nornir/badge.svg?branch=develop)](https://coveralls.io/github/nornir-automation/nornir?branch=develop)\n\n\nNornir\n=======\n\n![logo][logo]\n\nNornir is a pure Python automation framework intented to be used directly from Python. While most automation frameworks use their own Domain Specific Language (DSL) which you use to describe what you want to have done, Nornir lets you control everything from Python.\n\nOne of the benefits we want to highlight with this approach is the ease of troubleshooting, if something goes wrong you can just use your existing debug tools directly from Python (just add a line of `import pdb` & `pdb.set_trace()` and you\'re good to go). Doing the same using a DSL can be quite time consuming.\n\nWhat Nornir brings to the table is that it takes care of dealing with your inventory and manages the job of dispatching the tasks you want to run against your nodes and devices. The framework provides a very simple way to write plugins if you aren\'t happy with the ones we ship. Of course if you have written a plugin you think can be useful to others, please send us your code and test cases as a [pull request](https://github.com/nornir-automation/nornir/pulls).\n\n\nInstall\n=======\n\nPlease note that Nornir requires Python 3.6.2 or higher. Install Nornir with pip.\n\n```\npip install nornir\n```\n\nPlugins\n-------\n\nSince version 3.0.0 nornir doesn\'t ship with plugins, instead you can rely on `pip` to install them for you. You can find a non-exhaustive list of plugins in the following URL:\n\nhttps://nornir.tech/nornir/plugins/\n\nIf you wrote a plugin and want to add it to the list don\'t hesitate to [add it yourself](https://github.com/nornir-automation/nornir.tech.src/blob/master/data/nornir/plugins.yaml)\n\nDevelopment version\n-------------------\n\nIf you want to clone the repo and install it from there you will need to use [poetry](https://github.com/sdispater/poetry).\n\nDocumentation\n=============\n\nRead the [Nornir documentation](https://nornir.readthedocs.io/) online or review it\'s [code here](https://github.com/nornir-automation/nornir/tree/develop/docs)\n\nExamples\n========\n\nYou can find some examples and already made tools [here](https://github.com/nornir-automation/nornir-tools/)\n\nExternal Resources\n==================\n\nBelow you can find links to talks, blog posts, podcasts and other resources:\n\n* April 2019 - Packet Pushers podcast - [Heavy Networking 445: An Introduction To The Nornir Automation Framework](https://packetpushers.net/podcast/heavy-networking-445-an-introduction-to-the-nornir-automation-framework/)\n* May 2018 - Software Gone Wild podcast - [IPSpace podcast about nornir](http://blog.ipspace.net/2018/05/network-automation-with-brigade-on.html)\n* Sep 2018 - IPSpace network automation solutions - [Nornir workshop](https://my.ipspace.net/bin/list?id=NetAutSol&module=9#NORNIR) ([slides](https://github.com/dravetech/nornir-workshop/blob/master/nornir-workshop.pdf))\n* May 2018 - Networklore - [Introducing Nornir - The Python automation framework](https://networklore.com/introducing-brigade/)\n* May 2018 - Cisco blogs - [Exploring Nornir, the Python Automation Framework](https://blogs.cisco.com/developer/nornir-python-automation-framework)\n\n\nBugs & New features\n===================\n\nIf you think you have bug or would like to request a new feature, please register a GitHub account and [open an issue](https://github.com/nornir-automation/nornir/issues).\n\n\nContact & Support\n=================\n\nOfficial channel for communicating issues is via [GitHub issues](https://github.com/nornir-automation/nornir/issues) and you can use [GitHub discussions](https://github.com/nornir-automation/nornir/discussions) for general discussions around nornir. In addition, you can join the community in our ``#nornir`` channel in the [networktoCode Slack team](https://networktocode.herokuapp.com/).\n\n\nContributing to Nornir\n=======================\n\nIf you want to help the project, the [Contribution Guidelines](https://nornir.readthedocs.io/en/develop/contributing/index.html) is the best place to start.\n\n[logo]: docs/_static/logo/nornir_logo_02.jpg "nornir logo"\n',
    'author': 'David Barroso',
    'author_email': 'dbarrosop@dravetech.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nornir-automation/nornir',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Any, List

from nornir.core.inventory import Host


class F_BASE(object):
    def __call__(self, host: Host) -> bool:
        raise NotImplementedError()


class F_OP_BASE(F_BASE):
    def __init__(self, op1: F_BASE, op2: F_BASE) -> None:
        self.op1 = op1
        self.op2 = op2

    def __and__(self, other: F_BASE) -> "AND":
        return AND(self, other)

    def __or__(self, other: F_BASE) -> "OR":
        return OR(self, other)

    def __repr__(self) -> str:
        return "( {} {} {} )".format(self.op1, self.__class__.__name__, self.op2)


class AND(F_OP_BASE):
    def __call__(self, host: Host) -> bool:
        return self.op1(host) and self.op2(host)


class OR(F_OP_BASE):
    def __call__(self, host: Host) -> bool:
        return self.op1(host) or self.op2(host)


class F(F_BASE):
    def __init__(self, **kwargs: Any) -> None:
        self.filters = kwargs

    def __call__(self, host: Host) -> bool:
        return all(
            F._verify_rules(host, k.split("__"), v) for k, v in self.filters.items()
        )

    def __and__(self, other: "F") -> AND:
        return AND(self, other)

    def __or__(self, other: "F") -> OR:
        return OR(self, other)

    def __invert__(self) -> "F":
        return NOT_F(**self.filters)

    def __repr__(self) -> str:
        return "<Filter ({})>".format(self.filters)

    @staticmethod
    def _verify_rules(data: Any, rule: List[str], value: Any) -> bool:
        if len(rule) > 1:
            try:
                return F._verify_rules(data.get(rule[0], {}), rule[1:], value)
            except AttributeError:
                return False

        elif len(rule) == 1:
            operator = "__{}__".format(rule[0])
            if hasattr(data, operator):
                return bool(getattr(data, operator)(value))

            elif hasattr(data, rule[0]):
                if callable(getattr(data, rule[0])):
                    return bool(getattr(data, rule[0])(value))

                else:
                    return bool(getattr(data, rule[0]) == value)

            elif rule == ["in"]:
                return bool(data in value)
            elif rule == ["any"]:
                if isinstance(data, list):
                    return any([x in data for x in value])
                else:
                    return any([x == data for x in value])
            elif rule == ["all"]:
                if isinstance(data, list):
                    return all([x in data for x in value])
                else:
                    # it doesn't make sense to check a single value meets more than one case
                    return False
            else:
                return bool(data.get(rule[0]) == value)

        else:
            raise Exception(
                "I don't know how I got here:\n{}\n{}\n{}".format(data, rule, value)
            )


class NOT_F(F):
    def __call__(self, host: Host) -> bool:
        return not any(
            F._verify_rules(host, k.split("__"), v) for k, v in self.filters.items()
        )

    def __invert__(self) -> F:
        return F(**self.filters)

    def __repr__(self) -> str:
        return "<Filter NOT ({})>".format(self.filters)



from typing import Any, List, Type

from nornir.core.task import AggregatedResult, Task
from nornir.core.inventory import Host
from nornir.core.plugins.register import PluginRegister

from typing_extensions import Protocol


RUNNERS_PLUGIN_PATH = "nornir.plugins.runners"


class RunnerPlugin(Protocol):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        This method configures the plugin
        """
        raise NotImplementedError("needs to be implemented by the plugin")

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        """
        This method runs the given task over all the hosts
        """
        raise NotImplementedError("needs to be implemented by the plugin")


RunnersPluginRegister: PluginRegister[Type[RunnerPlugin]] = PluginRegister(
    RUNNERS_PLUGIN_PATH
)
